#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sqlite3.h>

#include "libbox-db.h"
#include "libbox-db-deps.h"
#include "libbox-db-utils.h"

static char box_db_mode_to_str(int m)
{
    if S_ISDIR(m)
        return 'D';
    else if S_ISCHR(m)
        return 'C';
    else if S_ISBLK(m)
        return 'B';
    else if S_ISFIFO(m)
        return 'P';
    else if S_ISLNK(m)
        return 'L';
#ifndef __BEOS__
    else if S_ISSOCK(m)
        return 'S';
#endif

    return 'F';
}

static void box_db_assign_file(sqlite3 *db, const gchar* path, gint package_id, gint last_id)
{
    gchar *sql = NULL;

    /* insert association */
    sql = g_strdup_printf("INSERT INTO file_list( package_id, file_id ) VALUES (%d, %d)", \
            package_id, last_id);
    box_db_execute_simple(db, sql);
    g_free(sql);
}

static gint box_db_get_or_create(sqlite3 *db, gchar *item, gint parent_id, gchar *symlink, gint package_id, int uid, int gid, int mode)
{
    sqlite3_stmt *stmt = NULL;
    int rc;
    gchar *sql = NULL;
    gint last_id = -1;
    gboolean any_result = FALSE;
    gchar *mode_str;


    /* check if file already exist */
    sql = g_strdup_printf("SELECT file_id FROM file WHERE name = '%s' AND parent_id = %d", item, parent_id);

    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            any_result = TRUE;
            last_id = sqlite3_column_int(stmt, 0);
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);
    g_free(sql);

    mode_str = g_strdup_printf("%o", mode);

    if (!symlink)
        symlink = g_strdup("");

    if (!any_result)
    {
        sql = g_strdup_printf("INSERT INTO file( file_id, parent_id, name, symlink, type, perms ) \
                VALUES (NULL, %d, '%s', '%s', '%c', '%s')", parent_id, item, symlink, box_db_mode_to_str(mode), mode_str);
        /* INSERT item */
        box_db_execute_simple(db, sql);
        last_id = sqlite3_last_insert_rowid(db);
        g_free(sql);

    } 

    if (strlen(symlink) == 0)
        g_free(symlink);

    g_free(mode_str);

    return last_id;
}

static gint box_db_add_file(sqlite3 *db, const gchar* path, gchar *symlink, int mode, gint package_id, int uid, int gid)
{
    gchar **tokens = NULL;
    gint i;
    gint parent_id = -1;

    tokens = g_strsplit(path, G_DIR_SEPARATOR_S, 0);
    for (i = 0; tokens != NULL && tokens[i] != NULL; i++)
    {
        parent_id = box_db_get_or_create(db, tokens[i], parent_id, symlink, package_id, uid, gid, mode);
    }
    g_strfreev(tokens);

    return parent_id;
}


static GList* box_db_dir_list_internal(sqlite3 *db, const gchar *path, const gchar *path_relative, gint package_id, gint last_id, GList *files)
{
    GDir *dir = NULL;
    const gchar *name = NULL;
    gchar *file = NULL;
    gchar *file2 = NULL;
    gchar *new_path = NULL;
    GError *error = NULL;
    gboolean empty_dir = TRUE;
    struct stat buf;
    gchar *symlink = NULL;
    gint l;

    new_path = g_build_filename(path, path_relative, NULL);
    dir = g_dir_open(new_path, 0, &error);
    g_free(new_path);

    if (dir == NULL)
    {
        if (error != NULL)
        {
            printf("error: %s\n", error->message);
        }
        return NULL;
    }

    while ((name = g_dir_read_name(dir)) != NULL)
    {
        empty_dir = FALSE;
        file = g_build_filename(path, path_relative, name, NULL);
        file2 = g_build_filename(path_relative, name, NULL);

        lstat(file, &buf);

        /* FIXME: possible buffer overflow */
        symlink = NULL;

        if (S_ISLNK(buf.st_mode)) {
            symlink = g_malloc0(4096);
            l = readlink(file, symlink, 4096);
        } 


        /*        printf("%s\t%d:%d,%o\n", file2, buf.st_uid, buf.st_gid, buf.st_mode); */

        files = g_list_append(files, g_strdup(file2));
        last_id = box_db_add_file(db, file2, symlink, buf.st_mode, package_id, buf.st_uid, buf.st_gid);

        g_free(symlink);

        if (!S_ISDIR(buf.st_mode))
        {
            box_db_assign_file(db, file2, package_id, last_id);
        } else
        {
            box_db_dir_list_internal(db, path, file2, package_id, last_id, files);
        }

        g_free(file2);
        g_free(file);
    }
    if (empty_dir)
    {
        box_db_assign_file(db, path_relative, package_id, last_id);
    }

    if (dir != NULL)
        g_dir_close(dir);

    return files;
}

GList* box_db_install(sqlite3 *db, const gchar *path, const gchar *path_relative, const gchar *name, const gchar *version, gint release, const gchar *summary)
{
    gchar *sql = NULL;
    GList *files = NULL;
    gint package_id = -1;
    gint packages_count;

    sql = g_strdup_printf("SELECT COUNT(*) FROM package WHERE name = '%s'", name);
    packages_count = box_db_execute_simple_query_get_int(db, sql);
    g_free(sql);

    if (packages_count == 0)
    {
        sql = sqlite3_mprintf("INSERT INTO package( package_id, name, version, 'release', inst_date, summary ) VALUES ( NULL, '%s', '%s', %d, CURRENT_TIMESTAMP, '%q')", name, version, release, summary);
        box_db_execute_simple(db, sql);
        sqlite3_free(sql);
    } else
    {
        sql = sqlite3_mprintf("UPDATE package SET version = '%q', 'release' = %d, inst_date = CURRENT_TIMESTAMP, summary = '%q' WHERE name = '%q'", version, release, summary, name);
        box_db_execute_simple(db, sql);
        sqlite3_free(sql);
    }

    sql = g_strdup_printf("SELECT package_id FROM package WHERE name = '%s'", name);
    package_id = box_db_execute_simple_query_get_int(db, sql);
    g_free(sql);

    /* remove files associations */
    sql = g_strdup_printf("DELETE FROM file_list WHERE package_id = %d", package_id);
    box_db_execute_simple(db, sql);
    g_free(sql);

    files = box_db_dir_list_internal(db, path, path_relative, package_id, -2, files);
    return files;
}


static void box_db_remove_file(sqlite3 *db, gint file_id, gboolean is_leaf)
{
    gchar *sql;
    gint children = 0;
    gint parent_id;

    sql = g_strdup_printf("SELECT parent_id FROM file WHERE file_id = %d", file_id);
    parent_id = box_db_execute_simple_query_get_int(db, sql);
    g_free(sql);

    sql = g_strdup_printf("SELECT COUNT(*) FROM file WHERE parent_id = %d", file_id);
    children = box_db_execute_simple_query_get_int(db, sql);
    g_free(sql);

    if (is_leaf || (children == 0))
    {
        sql = g_strdup_printf("DELETE FROM file WHERE file_id = %d", file_id);
        box_db_execute_simple(db, sql);
        g_free(sql);
        if (parent_id != -1)
            box_db_remove_file(db, parent_id, FALSE);
    }
}

GList* box_db_uninstall(sqlite3 *db, const gchar *package, gboolean force, GList **pkg_deps, gboolean *exist)
{
    gchar *sql, *sql2;
    int rc;
    sqlite3_stmt *stmt = NULL;
    gint file_id;
    gint ref_count = -1;
    gchar *file_str = NULL;
    GList *files = NULL;
    GList *dep_packages = NULL;
    gint package_id;

    sql = g_strdup_printf("SELECT package_id FROM package WHERE name = '%s'", package);
    package_id = box_db_execute_simple_query_get_int(db, sql);
    *exist = (package_id > 0);
    g_free(sql);

    /* remove dependencies */
    dep_packages = box_db_remove_deps(db, (gchar *)package, force);

    if (!force && (dep_packages != NULL))
    {
        *pkg_deps = dep_packages;
        return NULL;
    }

    sql = g_strdup_printf("SELECT file_id FROM file_list WHERE package_id = %d", package_id);
    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            file_id = sqlite3_column_int(stmt, 0);

            file_str = box_db_get_file_path_from_id(db, file_id, "main");
            files = g_list_prepend(files, file_str);

            sql2 = g_strdup_printf("SELECT COUNT(*) FROM file_list WHERE file_id = %d", file_id);
            ref_count = box_db_execute_simple_query_get_int(db, sql2);
            g_free(sql2);

            if (ref_count == 1)
            {
                box_db_remove_file(db, file_id, TRUE);
            }
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);
    g_free(sql);

    /* remove files associations */
    sql = g_strdup_printf("DELETE FROM file_list WHERE package_id = %d", package_id);
    box_db_execute_simple(db, sql);
    g_free(sql);


    /* remove package entry */
    sql = g_strdup_printf("DELETE FROM package WHERE package_id = %d", package_id);
    box_db_execute_simple(db, sql);
    g_free(sql);

    /* remove scripts */
    sql = g_strdup_printf("DELETE FROM script WHERE package_id = %d", package_id);
    box_db_execute_simple(db, sql);
    g_free(sql);

    return files;
}

gchar* box_db_get_file_path_from_id(sqlite3 *db, gint file_id, const gchar *db_prefix)
{
    gchar *sql = NULL;
    int rc;
    sqlite3_stmt *stmt = NULL;
    gchar *name = NULL;
    GList *list = NULL;
    GList *li = NULL;
    gchar *result = g_strdup("/");
    gchar *last_result = NULL;



    while (file_id > 0)
    {
        sql = g_strdup_printf("SELECT name, parent_id FROM %s.file WHERE file_id = %d", db_prefix, file_id);
        rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

        file_id = -1;
        name = NULL;
        while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
        {
            if (rc == SQLITE_ROW)
            {
                name = g_strdup((gchar *)sqlite3_column_text(stmt, 0));
                file_id = sqlite3_column_int(stmt, 1);
            } else if (rc != SQLITE_OK)
            {
                printf("ERROR: %s\n", sqlite3_errmsg(db));
            }
        }
        g_free(sql);
        list = g_list_append(list, name);
        sqlite3_finalize(stmt);
    }


    list = g_list_reverse(list);

    for (li = list; li != NULL; li = li->next)
    {
        last_result = result;
        result = g_build_filename(result, (gchar*)li->data, NULL);
        g_free(li->data);
        g_free(last_result);
    }

    g_list_free(list);

    return result;
}

GList* box_db_get_file_list(sqlite3 *db, const gchar *name)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    gint file_id;
    gint package_id;
    GList *files = NULL;
    gchar *file;
    gchar *sql = NULL;

    sql = g_strdup_printf("SELECT package_id FROM package WHERE name = '%s'", name);
    package_id = box_db_execute_simple_query_get_int(db, sql);
    g_free(sql);

    sql = g_strdup_printf("SELECT file_id FROM file_list WHERE package_id = %d", package_id);
    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            file_id = sqlite3_column_int(stmt, 0);
            file = box_db_get_file_path_from_id(db, file_id, "main");
            files = g_list_prepend(files, file);
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);
    g_free(sql);

    return files;
}

void box_db_file_list_free(GList *list)
{
    GList *li;

    for (li = list; li != NULL; li = li->next)
    {
        g_free(li->data);
    }
    g_list_free(list);
}


GList* box_db_search_file(sqlite3 *db, const gchar *name)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *files = NULL;
    gchar *sql = NULL;
    gint file_id;
    SearchResult *search_result = NULL;

    sql = g_strdup_printf("SELECT file_id, p.name FROM file_list f, package p WHERE p.package_id = f.package_id AND file_id IN (SELECT file_id FROM file WHERE name = '%s' AND type != 'D' ) ORDER BY p.name DESC", name);
    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            file_id = sqlite3_column_int(stmt, 0);

            search_result = g_new0(SearchResult, 1);
            search_result->file = box_db_get_file_path_from_id(db, file_id, "main");
            search_result->package = g_strdup((gchar *)sqlite3_column_text(stmt, 1));
            files = g_list_prepend(files, search_result);
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);

    g_free(sql);

    return files;
}

void box_db_search_result_free(SearchResult *result)
{
    g_free(result->file);
    g_free(result->package);
    g_free(result);
}

void box_db_search_result_list_free(GList *list)
{
    GList *li;

    for (li = list; li != NULL; li = li->next)
    {
        box_db_search_result_free((SearchResult*)li->data);
    }
    g_list_free(list);
}   

PackageInfo* box_db_package_info_new(const gchar *name, const gchar *version, const gchar *release, const gchar *inst_date)
{
    PackageInfo *result = NULL;

    result = g_new0(PackageInfo, 1);
    result->name = g_strdup(name);
    result->version = g_strdup(version);
    result->release = g_strdup(release);
    result->inst_date = g_strdup(inst_date);

    return result;
}

void box_db_package_info_free(PackageInfo *package)
{
    g_free(package->name);
    g_free(package->version);
    g_free(package->release);
    g_free(package->inst_date);
    g_free(package);
}

GList* box_db_get_installed_packages(sqlite3 *db)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *files = NULL;
    gchar *sql = NULL;
    PackageInfo *package = NULL;

    sql = g_strdup_printf("SELECT name, version, `release`, inst_date FROM package ORDER BY name DESC");
    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            package = box_db_package_info_new((gchar *)sqlite3_column_text(stmt, 0), (gchar *)sqlite3_column_text(stmt, 1), \
                    (gchar *)sqlite3_column_text(stmt, 2), (gchar *)sqlite3_column_text(stmt, 3));
            files = g_list_prepend(files, package);
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);

    g_free(sql);

    return files;
}

GList* box_db_search_dup_files(sqlite3 *db)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *files = NULL;
    gchar *sql = NULL;
    gint file_id;
    SearchResult *search_result = NULL;

    sql = g_strdup_printf("SELECT fl.file_id, p.name FROM file_list AS fl, package AS p, file AS f WHERE fl.file_id IN (SELECT file_id FROM file_list GROUP BY file_id HAVING COUNT(file_id) > 1) AND p.package_id = fl.package_id AND f.type != 'D' AND f.file_id = fl.file_id ORDER by fl.file_id ");
    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            file_id = sqlite3_column_int(stmt, 0);

            search_result = g_new0(SearchResult, 1);
            search_result->file = box_db_get_file_path_from_id(db, file_id, "main");
            search_result->package = g_strdup((gchar *)sqlite3_column_text(stmt, 1));
            files = g_list_prepend(files, search_result);

        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);

    g_free(sql);

    return files;
}

gint box_db_get_package_id(sqlite3 *db, const gchar* package)
{
    gint package_id;
    gchar *sql = NULL;

    sql = g_strdup_printf("SELECT package_id FROM package WHERE name = '%s'", package);
    package_id = box_db_execute_simple_query_get_int(db, sql);
    g_free(sql);

    return package_id;
}


void box_db_install_script(sqlite3 *db, gint package_id, const gchar *script, enum SCRIPT_TYPE type, gint status)
{
    gchar *sql = NULL;
    gint count;

    sql = g_strdup_printf("SELECT COUNT(*) FROM script WHERE package_id = %d AND type = %d",
        package_id, type);
    count = box_db_execute_simple_query_get_int(db, sql);
    g_free(sql);

    if (count == 0)
    {
        sql = sqlite3_mprintf("INSERT INTO script(package_id, type, content, status) VALUES \
            (%d, %d, '%q', %d)", package_id, type, script, status);
        box_db_execute_simple(db, sql);
        sqlite3_free(sql);
    } else
    {
        sql = sqlite3_mprintf("UPDATE script SET content = '%q', status = %d WHERE package_id = %d \
            AND type = %d", script, status, package_id, type);
        box_db_execute_simple(db, sql);
        sqlite3_free(sql);
    }

}


GList* box_db_get_scripts(sqlite3 *db, enum SCRIPT_TYPE script_type, gboolean run_new)
{
    gchar *sql = NULL;
    GList *list = NULL;

    if (run_new)
    {
        sql = g_strdup_printf("SELECT DISTINCT s.content FROM script s \
            WHERE s.type = %d AND s.status = 0", script_type);
    } else
    {
        sql = g_strdup_printf("SELECT DISTINCT s.content FROM script s \
            WHERE s.type = %d", script_type);
    }

    list = box_db_execute_simple_query_get_string_list(db, sql);
    

    g_free(sql);

    return list;
}
