#include <glib.h>
#include <glib/gstdio.h>
#include <sqlite3.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>

#include "config.h"
#include "libbox-db-utils.h"
#include "libbox-db-repos.h"
#include "libbox-db-deps.h"
#include "libbox-db.h"
#include "libbox.h"
#include "libbox-parser.h"
#include "libbox-download.h"

static int run_cmd(const gchar *cmd)
{
    /* g_spawn_command_line_sync(cmd, NULL, NULL, &exit, NULL); */
   return system(cmd);
}

static gint get_release_from_filename(const gchar *string)
{
    GRegex *regex;
    GMatchInfo *match_info;
    gint result;
    gchar *word;

    regex = g_regex_new ("-([0-9]+)\\.[^\\.]+\\.box$", 0, 0, NULL);
    g_regex_match (regex, string, 0, &match_info);

    word = g_match_info_fetch (match_info, 1);
    result = atoi(word);
    g_free (word);

    g_match_info_free (match_info);
    g_regex_unref (regex);

    return result;
}

static gchar* get_package_name_from_file_and_version(const gchar *string, const gchar *version, gint release)
{
    GRegex *regex;
    GMatchInfo *match_info;
    gchar *result;
    gchar *basename;
    gchar *r;
    gchar *word;
    gchar *version_escaped;

    basename = g_path_get_basename (string);
    version_escaped = g_regex_escape_string (version, -1);
    r = g_strdup_printf("^(.*)-%s-%d\\.[^\\.]+\\.box$", version_escaped, release);

    regex = g_regex_new (r, 0, 0, NULL);
    g_regex_match (regex, basename, 0, &match_info);

    word = g_match_info_fetch (match_info, 1);
    g_assert(word);
    result = g_strdup(word);
    g_free (word);

    g_free (r);
    g_free (version_escaped);
    g_free (basename);
    g_match_info_free (match_info);
    g_regex_unref (regex);
    
    return result;
}
static gchar* get_package_name_from_binary_file(const gchar *file)
{
    gchar *cmd;
    gchar *temp;
    BoxDocument *document;
    gchar *result;

    temp = g_build_filename(g_get_tmp_dir(), "box.XXXXXX", NULL);
    temp = (gchar*) mktemp((char*)temp);

    g_mkdir_with_parents(temp, 0755);

    /* unpack package */
    cmd = g_strdup_printf("tar xf %s -C %s package.box", file, temp);
    run_cmd(cmd);
    g_free(cmd);

    /* parse package.box */
    cmd = g_strdup_printf("%s/package.box", temp);
    document = box_document_new(cmd);
    g_free(cmd);
    result = g_strdup(document->header->name);
    box_document_free(document);

    /* clean */
    cmd = g_strdup_printf("rm -rf %s", temp);
    run_cmd(cmd);
    g_free(cmd);
   
    g_free(temp);

    return result;
}

static gchar* _find_script_str(gchar *pkgname, BoxDocument* document, gint type)
{
    gchar *script = NULL;
    GList *li;
    BoxPackage *package;

    for (li = document->packages; li != NULL; li = li->next)
    {
        package = (BoxPackage*)li->data;
        if (strcmp(pkgname, package->name) == 0 || strcmp("normal", package->name) == 0)
        {
            if (type == 0)
                script = package->pre_install;
            else if (type == 1)
                script = package->post_install;
        }
    }

    return script;
}

static gchar* _macro(const gchar* script)
{
    GRegex *regex;
    gchar *new_script = NULL;

    regex = g_regex_new ("%{([^}]+)}", 0, 0, NULL);
    new_script = g_regex_replace(regex, script, -1, 0, "`_macro \\1 $tmpdir_files`", G_REGEX_MATCH_NOTBOL, NULL);

    g_regex_unref(regex);

    return new_script;
}

/*
static gint _get_script_type_by_name(const gchar *name)
{
   gint type = -1;

   if (strcmp(name, "pre-install") == 0)
   {
       type = SCRIPT_TYPE_PRE_INSTALL;
   } else if (strcmp(name, "post-install") == 0)
   {
       type = SCRIPT_TYPE_POST_INSTALL;
   }

   return type;
}
*/

static const gchar* _get_script_name_by_type(enum SCRIPT_TYPE type)
{
    switch (type)
    {
        case SCRIPT_TYPE_PRE_INSTALL:
            return "pre-install";
        case SCRIPT_TYPE_POST_INSTALL:
            return "post-install";
        default:
            return NULL;
    }
};

static gchar* _get_script_str_original(gchar *pkgname, BoxDocument* document,
        const gchar *temp, enum SCRIPT_TYPE type)
{
    gchar *xml = NULL;
    gchar *script = NULL;
    gchar *file_script = NULL;
    gchar *file = NULL;

    xml = _find_script_str(pkgname, document, type);

    if (!xml)
        xml = g_strdup("");

    file = g_build_filename(temp, "cache", _get_script_name_by_type(type), NULL);

    if (!g_file_get_contents(file, &file_script, NULL, NULL))
        file_script = g_strdup("");

    g_free(file);

    if ((strlen(xml) + strlen(file_script)) > 0)
        script = g_strdup_printf("%s\n%s\n", xml, file_script);

    g_free(file_script);
    g_free(xml);

    return script;
}

static gchar* _get_script_str(const gchar *file_script, enum SCRIPT_TYPE script_type)
{
    gchar *result = NULL;
    gchar *script = NULL;

    if (script_type == SCRIPT_TYPE_PRE_INSTALL)
    {
        script = g_strdup_printf(
            "#!/bin/sh\n\n" \
            "export LC_ALL=C; export LANG=C;\n" \
            "export PATH="PACKAGE_PREFIX"/sbin:"PACKAGE_PREFIX"/sbin:/sbin:/bin:"PACKAGE_PREFIX"/bin\n\n" \
            ". "PACKAGE_LIBDIR"/box/macros\n" \
            "%s\n", file_script);
    } else if (script_type == SCRIPT_TYPE_POST_INSTALL)
    {
        script = g_strdup_printf(
            "#!/bin/sh\n\n" \
            "tmpdir_files=$1\n" \
            "export LC_ALL=C; export LANG=C;\n" \
            "export PATH="PACKAGE_PREFIX"/sbin:"PACKAGE_PREFIX"/sbin:/sbin:/bin:"PACKAGE_PREFIX"/bin\n\n" \
            "export PATH=\"$tmpdir_files/bin:$tmpdir_files"PACKAGE_PREFIX"/bin:$PATH\"\n" \
            "export LD_LIBRARY_PATH=\"$tmpdir_files/lib:$tmpdir_files"PACKAGE_PREFIX"/lib$LD_LIBRARY_PATH\"\n\n" \
            "export LIBRARY_PATH=\"$tmpdir_files/lib:$tmpdir_files/boot/common/lib$LIBRARY_PATH\"\n\n" \
            ". "PACKAGE_LIBDIR"/box/macros\n" \
            "%s\n", file_script);
    }

    result = _macro(script);

    g_free(script);

    return result;
}


static void _run_script(const gchar *script, const gchar *tmpdir, enum SCRIPT_TYPE type)
{
    gchar *tempfile;
    gchar *filesdir;
    gchar *cmd;

    tempfile = g_build_filename(tmpdir, _get_script_name_by_type(type), NULL);
    filesdir = g_build_filename(tmpdir, "data", NULL);

    g_file_set_contents(tempfile, script, -1, NULL);

    cmd = g_strdup_printf("sh %s %s", tempfile, filesdir);
    g_free(filesdir);

    run_cmd(cmd);
    g_free(cmd);
    
    g_free(tempfile);
}

static gchar* _get_summary(const gchar *temp)
{
    GError *error = NULL;
    gchar *tmpstr;
    gchar *summary;
    gint i = 0;

    tmpstr = g_strdup_printf("%s/cache/summary", temp);
    if (!g_file_get_contents(tmpstr, &summary, NULL, &error))
        summary = g_strdup("");
    else {
        for (i=0; i<strlen(summary); i++)
        {
            if (summary[i] == 0x0a)
                summary[i] = 0x20;
            g_strstrip(summary);
        }
    }
    g_free(tmpstr);

    return summary;
}

static void
_remove_old_files(GList *files0, GList *files, const gchar *dest_dir)
{
    GList *li0, *li;
    gchar *full_path;

    for (li0 = files0; li0 != NULL; li0 = li0->next)
    {
        gboolean exist = FALSE;

        for (li = files; li != NULL; li = li->next)
        {
            if (strcmp((gchar *)li0->data, (gchar *)li->data) == 0)
                exist = TRUE;
        }
        if (!exist)
        {
            full_path = g_build_filename(dest_dir, (gchar *)li0->data, NULL);
            g_unlink(full_path);
            g_free(full_path);
        }
    }
    for (li = files; li != NULL; li = li->next)
    {
        full_path = g_build_filename(dest_dir, (gchar *)li->data, NULL);
        g_free(full_path);
    }
}
static void
_set_permissions(const gchar *temp, const gchar *dest_dir)
{
    gint i = 0, j = 0;
    gchar *tmpstr;
    gchar *file_buffer = NULL;
    gchar **lines = NULL;
    gchar **fields = NULL;
    gchar *tmpstr2 = NULL;
    gchar *cmd;
    gint perms;
    struct stat buf;

    tmpstr = g_strdup_printf("%s/cache/perms", temp);
    if (g_file_get_contents(tmpstr, &file_buffer, NULL, NULL))
    {
        lines = g_strsplit(file_buffer, "\n", 0);
        g_free(file_buffer);

        while (lines[i])
        {
            j = 0;
            fields = g_strsplit(lines[i], "\t", 0);
            while (fields[j])
            {
                if (j == 1) {
                    tmpstr2 = g_build_filename (dest_dir, fields[j], NULL);
                    if (g_file_test( tmpstr2, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR))
                    {
                        g_lstat(tmpstr2, &buf);
                        cmd = g_strdup_printf("chown %s %s", fields[0], tmpstr2);
                        run_cmd(cmd);
                        g_free(cmd);
                        if (fields[2]) {
                            /* set new permissions
                               use external chmod here because permission can be in text format
                               like "a=r" */
                            cmd = g_strdup_printf("chmod %s %s", fields[2], tmpstr2);
                            run_cmd(cmd);
                            g_free(cmd);
                        } else {
                            /* restore permissions after chown */
                            perms = buf.st_mode;
                            g_chmod(tmpstr2, perms);
                        }
                    }
                    g_free(tmpstr2);
                }
                j++;
            }
            i++;
        }
        g_strfreev(lines);
    }
    g_free(tmpstr);
}

static void
cmd_printf(const gchar *format, ...)
{
    gchar *cmd;
    va_list args;

    va_start (args, format);
    cmd = g_strdup_vprintf(format, args);
    va_end (args);

    run_cmd(cmd);
    g_free(cmd);
}

static void print_list_str(GList *list)
{
    GList *li = NULL;
    PackageSearch *search = NULL;

    for (li = list; li != NULL; li = li->next)
    {
        search = (PackageSearch *)li->data;
        printf("- %s\n", search->package);
    }
}

static gboolean
_box_package_install_file(const gchar *filename, const gchar *dest_dir, progressFunc progress, gpointer user_data, sqlite3 *db_repos, gboolean install_files, gboolean no_deps)
{
    gchar *temp;
    BoxDocument *document;
    sqlite3 *db;
    gchar *arg_pkgname = NULL;
    GList *files = NULL;
    GList *files0 = NULL;
    gchar *tmpstr;
    gchar *summary;
    gint release;
    gint package_id;
    gchar *pre_install_orig = NULL;
    gchar *post_install_orig = NULL;
    gchar *pre_install = NULL;
    gchar *post_install = NULL;
    GList *deps;
    gboolean abort = FALSE;
    gboolean native;
    gint status;

    if (!g_file_test(filename, G_FILE_TEST_EXISTS))
        return FALSE;

    temp = g_build_filename(g_get_tmp_dir(), "box-install.XXXXXX", NULL);
    temp = (gchar*) mktemp((char*)temp);

    g_mkdir_with_parents(temp, 0755);

    /* unpack package */
    cmd_printf("tar xf %s -C %s", filename, temp);

    /* unpack data */
    cmd_printf("tar xfz %s/data/*.tar.gz -C %s/data/", temp, temp);

    /* clean unused file */
    cmd_printf("rm -f %s/data/*.tar.gz", temp);

    if (strcmp(dest_dir, "/") == 0)
        native = TRUE;
    else
        native = FALSE;


    /* ------------ BEGIN DB TRANSACTION ------------- */
    db = box_db_open(dest_dir);
    box_db_create_empty(db);
    box_db_transaction_begin(db);

    /* parse package.box */
    tmpstr = g_strdup_printf("%s/package.box", temp);
    document = box_document_new(tmpstr);
    g_free(tmpstr);

    release = get_release_from_filename(filename);

    arg_pkgname = get_package_name_from_file_and_version(filename, document->header->version, release);
   
    /* read files list from DB */
    files0 = box_db_get_file_list(db, arg_pkgname);

    /* install files from filesystem to the DB */
    summary = _get_summary(temp);
    tmpstr = g_strdup_printf("%s/data", temp);

    files = box_db_install(db, tmpstr, "/", arg_pkgname, document->header->version, release, summary);
    g_free(tmpstr);
    g_free(summary);

    package_id = box_db_get_package_id(db, arg_pkgname);

    tmpstr = g_strdup_printf("%s/cache/provides", temp);
    /* check dependencies if upgrading */
    if (!no_deps)
    {
        deps = box_db_check_deps_upgrade(db_repos, tmpstr, arg_pkgname);
        if (deps)
        {
            printf("Dependencies problem:\n");
            print_list_str(deps);
            box_db_transaction_rollback(db);
            abort = TRUE;
        }
        box_db_repos_package_list_free(deps);
    }

    if (!abort)
    {
        /* install dependencies */
        box_db_install_provides(db, package_id, tmpstr);
        g_free(tmpstr);

        tmpstr = g_strdup_printf("%s/cache/requires", temp);
        box_db_install_requires(db, package_id, tmpstr);
        g_free(tmpstr);

        tmpstr = g_strdup_printf("%s/cache/buildrequires", temp);
        box_db_install_build_requires(db, package_id, tmpstr);
        g_free(tmpstr);

        tmpstr = g_strdup_printf("%s/cache/buildprovides", temp);
        box_db_install_build_provides(db, package_id, tmpstr);
        g_free(tmpstr);

        /* read original scripts */
        pre_install_orig = _get_script_str_original(arg_pkgname, document, temp, SCRIPT_TYPE_PRE_INSTALL);
        post_install_orig = _get_script_str_original(arg_pkgname, document, temp, SCRIPT_TYPE_POST_INSTALL);

        /* save original scripts into db */
        status = 0;
        if (native)
            status = 1;
        
        if (pre_install_orig)
            box_db_install_script(db, package_id, pre_install_orig, SCRIPT_TYPE_PRE_INSTALL, status);

        if (post_install_orig)
            box_db_install_script(db, package_id, post_install_orig, SCRIPT_TYPE_POST_INSTALL, status);
    }

    /* -------------- END DB TRANSACTION ---------------- */
    if (!abort)
        box_db_transaction_end(db);
    box_db_close(db);

    if (!abort && install_files)
    {
        /* run pre-install script */
        pre_install = _get_script_str(pre_install_orig, SCRIPT_TYPE_PRE_INSTALL);
        if (pre_install_orig)
        {
            if (native)
            {
                _run_script(pre_install, temp, SCRIPT_TYPE_PRE_INSTALL);
            }
            g_free(pre_install_orig);
        }
        g_free(pre_install);

        /* remove files from filesystem which doesn't exist in a new pacakge */
        _remove_old_files(files0, files, dest_dir);
        box_db_file_list_free(files0);

        /* remove old files from the root */
        cmd_printf("cd %s/data; find . -not -type d -exec rm -f %s{} \\;", temp, dest_dir);

        /* install files to the real root */
        cmd_printf("cd %s/data; tar cf - * | tar xf - -C %s", temp, dest_dir);

        /* set correct file permissions */
        _set_permissions(temp, dest_dir);

        /* run post-install script */
        post_install = _get_script_str(post_install_orig, SCRIPT_TYPE_POST_INSTALL);
        if (post_install_orig)
        {
            if (native)
            {
                _run_script(post_install, temp, SCRIPT_TYPE_POST_INSTALL);
            }
            g_free(post_install_orig);
        }
        g_free(post_install);
    }

    box_document_free(document);

    /* clean */
    cmd_printf("rm -rf %s", temp);

    g_free(temp);
    g_free(arg_pkgname);

    return !abort;
}

const
gchar* box_get_repos_dir(const gchar *dest_dir)
{
    const gchar *env;

    env = g_getenv("BOX_USE_DEST_DB");
    
    if (env)
        return dest_dir;
    else
        return "/";
}

static gboolean
box_package_install_file(const gchar *filename, const gchar *dest_dir, progressFunc progress, gpointer user_data, sqlite3 *db_repos, gboolean no_deps)
{
    return _box_package_install_file(filename, dest_dir, progress, user_data, db_repos, TRUE, no_deps);
}

gboolean
box_package_install_factory(const gchar *filename, const gchar *dest_dir, progressFunc progress, gpointer user_data)
{
    sqlite3 *db_repos;
    gboolean result;

    db_repos = box_db_open(dest_dir);
    box_db_attach_repos(db_repos, box_get_repos_dir(dest_dir) );
    box_db_repos_init(db_repos);

    result = _box_package_install_file(filename, dest_dir, progress, user_data, db_repos, FALSE, FALSE);

    box_db_detach_repos(db_repos);
    box_db_close(db_repos);

    return result;
}

gboolean box_upgrade_dist(const gchar *dest_dir, progressFunc progress, gpointer user_data)
{
    GList *list;
    sqlite3* db = NULL;
    GList *li = NULL;
    PackageSearch *search = NULL;

    db = box_db_open(dest_dir);
    box_db_attach_repos(db, box_get_repos_dir(dest_dir));
    box_db_repos_init(db);

    list = box_db_repos_packages_for_upgrade(db);
   
    box_db_detach_repos(db);
    box_db_close(db);

    for (li = list; li != NULL; li = li->next)
    {
        search = (PackageSearch *)li->data;
        box_package_install(search->package, dest_dir, progress, user_data, FALSE);
    }

    box_db_repos_package_list_free(list);

    return TRUE;
}

static gboolean
_box_package_install(const gchar *package, const gchar *dest_dir, progressFunc progress, gpointer user_data, sqlite3 *db, gboolean result, gboolean no_deps)
{
    GList *list;
    GList *li;
    PackageSearch *search = NULL;
    gchar *filename, *out;

    list = box_db_repos_package_get_for_installation(db, (gchar*)package);

    li = g_list_first(list);
    if (li)
    {
        search = (PackageSearch *)li->data;
        filename = g_strdup(search->location);

        /* real installation */
        printf("Installing %s..., %s\n", package, filename);

        out = g_path_get_basename(filename);

        if (box_download_file(filename, out))
        {
            result &= box_package_install_file(out, dest_dir, progress, user_data, db, no_deps);
            g_unlink(out);
        } else {
            result = FALSE;
        }

        g_free(filename);
        g_free(out);
    } else
    {
        result = FALSE;
    }
    box_db_repos_package_list_free(list);

    return result;
}

gboolean box_package_install(const gchar *package, const gchar *dest_dir, progressFunc progress, gpointer user_data, gboolean no_deps)
{
    DepsResolve *deps;
    sqlite3* db = NULL;
    GList *li = NULL;
    gboolean result = TRUE;
    gboolean local_file = FALSE;
    gchar *filename = NULL;
    gchar *pkg;

    if (g_file_test(package, G_FILE_TEST_EXISTS) &&
        g_str_has_suffix(package, ".box") )
    {
        filename = g_strdup(package);
        pkg = get_package_name_from_binary_file(filename);
        local_file = TRUE;
    } else {
        pkg = g_strdup(package);
    }

    db = box_db_open(dest_dir);
    box_db_create_empty(db);
    box_db_attach_repos(db, box_get_repos_dir(dest_dir) );
    box_db_repos_init(db);

    deps = box_db_repos_resolve_deps_recurse(db, pkg, TRUE);

    if (deps->deps)
    {
        printf("Also will be installed: ");
        for (li = deps->deps; li != NULL; li = li->next)
        {
            printf(" %s", (gchar*)li->data);
        }
        printf("\n");
    }

    if (!no_deps)
    {
        if (deps->missing)
        {
            printf("Missing dependencies for %s:", package);

            for (li = deps->missing; li != NULL; li = li->next)
            {
                printf(" %s", (gchar*)li->data);
            }
            printf("\n");

            box_db_detach_repos(db);
            box_db_close(db);
            g_free(pkg);
            return FALSE;
        }
    }

    for (li = deps->deps; li != NULL; li = li->next)
    {
        result = _box_package_install((gchar *)li->data, dest_dir, progress, user_data, db, result, no_deps);
        if (!result)
        {
            printf("Cannot install %s!\n", (gchar *)li->data);
            break;
        }
    }
    box_db_repos_deps_free(deps);

    if (result) {
        if (local_file) {
            printf("Installing %s..., %s\n", pkg, filename);
            result = box_package_install_file(filename, dest_dir, progress, user_data, db, no_deps);
           g_free(filename);
        } else {
            result = _box_package_install(pkg, dest_dir, progress, user_data, db, result, no_deps);
        }
    }
    else
        printf("Cannot install %s!\n", package);

    if (!result)
        printf("Installation FAILED!\n");

    box_db_detach_repos(db);
    box_db_close(db);
    g_free(pkg);

    return result;
}

gboolean box_package_uninstall(const gchar *package_name, const gchar *dest_dir, progressFunc progress, gpointer user_data, gboolean force)
{
    sqlite3* db = NULL;
    GList *pkg_deps = NULL;
    GList *files;
    GList *li;
    gchar *full_path;
    float prog = 10.0;
    float step = 0.0;
    gboolean exist = FALSE;
    gchar *msg;

    db = box_db_open(dest_dir);

    box_db_transaction_begin(db);
    files = box_db_uninstall(db, package_name, force, &pkg_deps, &exist);
    box_db_transaction_end(db);

    if (exist == FALSE) {
        box_db_close(db);
        printf("Package %s doesn't exist.\n", package_name);
        return FALSE;
    }

    progress((int) prog, user_data);

    if (!force && (pkg_deps != NULL))
    {
        printf("Cannot uninstall %s, because some packages depends on it:\n", package_name);

        msg = box_glist_join_printf(pkg_deps, ", ", "%s");
        printf("%s\n", msg);
        g_free(msg);

        box_db_file_list_free(files);
        box_db_close(db);

        progress(100, user_data);

        return FALSE;
    }

    step = 89.9 / g_list_length(files);

    for (li = files; li != NULL; li = li->next)
    {
        full_path = g_build_filename(dest_dir, (gchar *)li->data, NULL);
        printf("%s\n", full_path);
        g_remove(full_path);
        g_free(full_path);

        prog += step;
        progress(prog, user_data);
    }

    box_db_file_list_free(files);
    box_db_close(db);

    progress(100, user_data);

    return TRUE;
}

gchar *box_get_repos_conf()
{
    const gchar *var;

    var = g_getenv("BOX_REPOS_FILE");
    if (var)
        return g_strdup(var);

    return g_strdup(_DEFAULT_REPOS_FILE);
}

gchar *box_get_db_dir()
{
    const gchar *var;

    var = g_getenv("BOX_DB_DIR");
    if (var)
        return g_strdup(var);

    return g_strdup(_DEFAULT_DB_DIR);
}

gchar *box_get_db_file()
{
    gchar *dir;
    gchar *result;

    dir = box_get_db_dir();

    result = g_build_filename(dir, "database", NULL);

    g_free(dir);

    return result;
}

gchar*
box_glist_join_printf(GList *list, const gchar *join, const char *printf_str)
{
    GList *li;
    gint n;
    gchar **str_array = NULL;
    gchar *result;

    n = g_list_length(list);
    str_array = g_new (gchar*, n + 1);
    str_array[n--] = NULL;

    for (li = list; li != NULL; li = li->next)
        str_array[n--] = g_strdup_printf(printf_str, (gchar*)li->data);

    result = g_strjoinv(join, str_array);
    g_strfreev(str_array);

    return result;
}



void
box_run_scripts(const gchar* dest_dir, gint script_type, gboolean run_new)
{
    sqlite3* db = NULL;
    GList *scripts = NULL;
    GList *li = NULL;
    gchar *tmpdir = NULL;
    gchar *script = NULL;

    db = box_db_open(dest_dir);

    tmpdir = g_build_filename(g_get_tmp_dir(), "box-scripts.XXXXXX", NULL);
    tmpdir = (gchar*) mktemp((char*)tmpdir);
    g_mkdir_with_parents(tmpdir, 0755);

    box_db_transaction_begin(db);

    scripts = box_db_get_scripts(db, script_type, run_new);

    for (li = scripts; li != NULL; li = li->next)
    {
        script = _get_script_str(li->data, script_type);
        _run_script(script, tmpdir, script_type);
        g_free(script);
    }

    /* TODO: free memory - scripts */

    /* clean */
    cmd_printf("rm -rf %s", tmpdir);

    g_free(tmpdir);

    box_db_transaction_end(db);
}
